@@ -5,7 +5,7 @@
 ;; Author: Adam Porter <adam@alphapapa.net>
 ;; Maintainer: Adam Porter <adam@alphapapa.net>
 ;; URL: https://github.com/alphapapa/plz.el
-;; Version: 0.9
+;; Version: 0.9.1
 ;; Package-Requires: ((emacs "27.1"))
 ;; Keywords: comm, network, http
 
@@ -359,7 +359,8 @@ It may be:
 - `(file FILENAME)' to pass FILENAME after having saved the
   response body to it without decoding.  FILENAME must be a
   non-existent file; if it exists, it will not be overwritten,
-  and an error will be signaled.
+  and an error will be signaled.  FILENAME is passed through
+  `expand-file-name', which see.
 
 - A function, which is called in the response buffer with it
   narrowed to the response body (suitable for, e.g. `json-read').
@@ -454,7 +455,9 @@ into the process buffer.
                                                 (`(file ,(and (pred stringp) as-filename))
                                                  (when (file-exists-p as-filename)
                                                    (error "File exists, will not overwrite: %S" as-filename))
-                                                 (setf filename as-filename)
+                                                 ;; Use `expand-file-name' because curl doesn't
+                                                 ;; expand, e.g. "~" into "/home/...".
+                                                 (setf filename (expand-file-name as-filename))
                                                  (list (cons "--output" filename))))))
                                      ((or 'put 'post)
                                       (append (list (cons "--dump-header" "-")
@@ -466,7 +469,9 @@ into the process buffer.
                                                 (`(file ,(and (pred stringp) as-filename))
                                                  (when (file-exists-p as-filename)
                                                    (error "File exists, will not overwrite: %S" as-filename))
-                                                 (setf filename as-filename)
+                                                 ;; Use `expand-file-name' because curl doesn't
+                                                 ;; expand, e.g. "~" into "/home/...".
+                                                 (setf filename (expand-file-name as-filename))
                                                  (list (cons "--output" filename))))
                                               (list
                                                ;; It appears that this must be the last argument
@@ -487,7 +492,9 @@ into the process buffer.
                                                 (`(file ,(and (pred stringp) as-filename))
                                                  (when (file-exists-p as-filename)
                                                    (error "File exists, will not overwrite: %S" as-filename))
-                                                 (setf filename as-filename)
+                                                 ;; Use `expand-file-name' because curl doesn't
+                                                 ;; expand, e.g. "~" into "/home/...".
+                                                 (setf filename (expand-file-name as-filename))
                                                  (list (cons "--output" filename))))))
                                      ('head
                                       (list (cons "--head" "")
